#!/usr/sbin/perl -w
#
# perlRat.pl v1.0.9a
# Modified April 29 17, 2004 11:49
# (c) Olaf R.P. Bininda-Emonds
#
# Generates:
#   A nexus-formatted paup block with randomly selected characters and
#   necessary instructions for performing a parsimony ratchet
#
# Usage: perlRat.pl -b<nodes> -c<characters> or -f<filename> [-l] [-n<batch.replicates>] [-p<percent>] [-s] [-t<constraint tree name>] [-w<factor>] [-x<wtset name>] [-v] [-h]
#   options: -b<nodes> = number of nodes to search for Bremer decay indices
#            -c<characters> = number of characters in data matrix
#            -f<filename> = name of NEXUS-formatted data file
#            -l = log ratchet replicates to disk
#            -n<batch.replicates> = number of ratchet replicates in the form x batches of
#                                   y replicates each
#                                   (default = 5.200)
#            -p<percent> = percent of characters to be upweighted in each step
#                          (default = 25)
#            -s = perform final brute force search on ratchet trees
#            -t<constraint tree name> = perform searches under topological constraints
#            -w<factor> = factor by which to upweight selected characters
#                         (default = 2)
#            -x<wtset name> = name of wtset (specified in previous assumptions block)
#							  containing original weights (default wtset = 1:all)
#            -v = verbose output
#            -h = print this message and quit

use strict;

# Set global defaults
	my $NumBremer = 1;
	my $NumChar = 0;
		my @MasterCharString;
	my $filename;
	my $log = 0;
	my $NumBatch = 5;
	my $NumRepl = 200;
	my $PercIncl = 0.25;
	my $search = 0;
	my $constraintTree;
	my $weight = 2;
	my ($wtfile, $origWeight);
	my $verbose = 0;
	
	my $date = scalar (localtime);

for (my $i = 0; $i <= $#ARGV; $i++)
	{
	if ($ARGV[$i] =~ /-b([\d]+)/)
		{
		$NumBremer = $1;
		}
	elsif ($ARGV[$i] =~ /-c([\d]+)/)
		{
		$NumChar = $1;
		}
	elsif ($ARGV[$i] =~ /-f([\w+\.?\w+?]+)/)
		{
		$filename = $1;
		}
	elsif ($ARGV[$i] =~ /-n([\d]+)\.([\d]+)/)
		{
		$NumBatch = $1;
		$NumRepl = $2;
		}
	elsif ($ARGV[$i] =~ /-l/)
		{
		$log = 1;
		}
	elsif ($ARGV[$i] =~ /-p([\d]+)/)
		{
		$PercIncl = $1;
		$PercIncl /= 100;
		}
	elsif ($ARGV[$i] =~ /-s/)
		{
		$search = 1;
		}
	elsif ($ARGV[$i] =~ /-t(.+)/)
		{
		$constraintTree = $1;
		}
	elsif ($ARGV[$i] =~ /-w(\d+\.*\d*)/)
		{
		$weight = $1;
		}
	elsif ($ARGV[$i] =~ /-x(\S+)/)
		{
		$origWeight = $1;
		}
	elsif ($ARGV[$i] =~ /-v/)
		{
		$verbose = 1;
		}
	elsif ($ARGV[$i] eq "-h")
		{
		print "Usage: perlRat.pl -b<nodes> -c<number> or -f<filename> [-l] [-n<batch.replicates>] [-p<number>] [-s] [-t<constraint tree name>] [-w<number>] [-x<wtset name>] [-v] [-h]\n";
		print "Options: -b<nodes> = number of nodes to search for Bremer decay indices\n";
		print "         -c<characters> = number of characters in data matrix\n";
		print "         -f<filename> = name of NEXUS-formatted data file\n";
		print "         -n<batch.replicates> = number of ratchet replicates in the form x batches of y replicates\n";
		print "                                (default = 5.200)\n";
		print "         -p<percent> = percent of characters to be upweighted in each step\n";
		print "                       (default = 25)\n";
		print "         -s = perform final brute force search on ratchet trees\n";
		print "         -t<constraint tree name> = perform searches under topological constraints\n";
		print "         -w<factor> = factor by which to upweight selected characters\n";
		print "                      (default = 2)\n";
		print "         -x<wtset name> = name of wtset (specified in previous assumptions block)\n";
		print "                        containing original weights (default wtset = 1:all)\n";
		print "         -v = verbose output\n";
		print "         -h = print this message and quit\n";
		exit(0);
		}
	else
		{
		print "Don't understand argument: $ARGV[$i]\n";
		print "Usage: perlRat.pl -b<nodes> -c<number> or -f<filename> [-l] [-n<batch.replicates>] [-p<percent>] [-s] [-t<constraint tree name>] [-w<factor>] [-x<wtset name>] [-v] [-h]\n";
		exit(1); 
		}
	}

die "Must specify total number of characters or filename\n" if (not $NumChar and not $filename);
if ($filename)	#Get number of character from file (if specified)
	{
	setLineBreak($filename);
	open (fDATA, "<$filename") or die "Cannot open data file $filename\n";
	while (<fDATA>)
		{
		chomp;
		if ($_ =~ /nchar\s*=\s*(\d+)/i)
			{
			$NumChar = $1;
			}
		}
	}

die "Error reading number of characters" if (not $NumChar);
	my $NumIncl = int($NumChar * $PercIncl);
	
	for (1..$NumChar)		#Create array containing numbers of all characters
		{
		push @MasterCharString, $_;
		}

if ($origWeight)
	{
	$origWeight = "assume wtset = ".$origWeight.";";
	}
else
	{
	$origWeight = "weights 1:all;";
	}

open (fCMD, ">Ratchet.cmd") or die "Cannot open batch file Ratchet.cmd\n";
	print fCMD "[Ratchet script created on $date using perlRat.pl v.1.0.9\n\n";
	print fCMD " Ratchet parameters: NumChar = $NumChar\n";
	print fCMD "                     Inclusion Percentage = $PercIncl ($NumIncl characters)\n";
	print fCMD "                     Batches = $NumBatch\n";
	print fCMD "                     Replicates = $NumRepl\n";
	print fCMD "                     Enforce constraints = ";
	if ($constraintTree)
		{
		print fCMD "yes\n";
		}
	elsif ($NumBremer > 1)
		{
		print fCMD "yes (Bremer decay index)\n";
		}
	else
		{
		print fCMD "no\n";
		}
	print fCMD "                     Final search = ";
	if ($search)
		{
		print fCMD "yes]\n\n";
		}
	else
		{
		print fCMD "no]\n\n";
		}
	print fCMD "begin paup;\n";
	print fCMD "\tset notifybeep=no autoclose=yes";
	print fCMD " WarnRedef=no status=no" if ($NumBremer > 1);
	print fCMD ";\n";
	print fCMD "\tlog file=ratchet.log replace;\n" if ($log);

	my $TreeRepl = 0;
	my $maxTrees;
	while ($TreeRepl < $NumBremer)
		{
		$TreeRepl++;
		print fCMD "\t[!][!*** Constraint tree #$TreeRepl ***]\n" if ($NumBremer > 1);

		for (my $i = 1; $i <= $NumBatch; $i++)
			{
			my $PAUPseed=int(10000*rand());
			print fCMD "\n\t[!][!*** Batch $i of $NumBatch ***]\n" if ($NumBatch > 1);
			print fCMD "\n\t[!][!*** Replicate 0 (initial tree) ***]\n\n";
			print fCMD "\t$origWeight\n";
			print fCMD "\thsearch / addseq=random nreps=1 rseed=$PAUPseed swap=TBR multrees=no";
				if ($constraintTree)
					{
					print fCMD " enforce=yes constraints=$constraintTree;\n";
					}
				elsif ($NumBremer > 1)
					{
					print fCMD " enforce=yes converse constraints=AD$TreeRepl;\n";
					}
				else
					{
					print fCMD ";\n";
					}
			print fCMD "\tsavetrees file=ratchet.tre format=altnexus ";
			if ($i == 1)
				{
				print fCMD "replace;\n\n";
				}
			else
				{
				print fCMD "append;\n\n";
				}
		
			for (1..$NumRepl)
				{
				my @InclList;
			
				my @CharString = @MasterCharString;
				
				for (1..$NumIncl)		#Randomly select required proportion of characters from array without replacement
					{
					push @InclList, splice(@CharString,rand @CharString,1);
					}
					
				@InclList = sort {$a <=> $b} @InclList;
			
				print fCMD "\t[!][!*** Replicate #$_ ***]\n\n";
				print fCMD "\tweights $weight: @InclList;\n";
				print fCMD "\thsearch start=current swap=TBR multrees=no";
					if ($constraintTree)
						{
						print fCMD " enforce=yes constraints=$constraintTree;\n";
						}
					elsif ($NumBremer > 1)
						{
						print fCMD " enforce=yes converse constraints=AD$TreeRepl;\n";
						}
					else
						{
						print fCMD ";\n";
						}
				print fCMD "\t$origWeight\n";
				print fCMD "\thsearch start=current swap=TBR multrees=no";
					if ($constraintTree)
						{
						print fCMD " enforce=yes constraints=$constraintTree;\n";
						}
					elsif ($NumBremer > 1)
						{
						print fCMD " enforce=yes converse constraints=AD$TreeRepl;\n";
						}
					else
						{
						print fCMD ";\n";
						}
				print fCMD "\tsavetrees file=ratchet.tre format=altnexus append;\n\n";
				}
			}
		unless ($NumBremer > 1)
			{
			print fCMD "\tlog file=ratchet.log replace;\n\n" if (not $log);
			}	
		if ($NumBatch * ($NumRepl + 1) < 10000)
			{
			$maxTrees = 10000;
			}
		else
			{
			$maxTrees = $NumBatch * ($NumRepl + 1);
			}
		print fCMD "\tset maxtrees=$maxTrees increase=no autoclose=yes notifybeep=no tcompress=yes torder=left criterion=parsimony status=no errorbeep=no outroot=monophyl;\n\n";
		if ($search)
			{
			print fCMD "\t[!][!*** Final brute force search ***]\n\n";
			print fCMD "\t\tgettrees file=ratchet.tre allblocks=yes warntree=no;\n";
			print fCMD "\t\t$origWeight\n";
			print fCMD "\t\thsearch start=current swap=TBR multrees=yes";
				if ($constraintTree)
					{
					print fCMD " enforce=yes constraints=$constraintTree;\n";
					}
				elsif ($NumBremer > 1)
					{
					print fCMD " enforce=yes converse constraints=AD$TreeRepl;\n";
					}
				else
					{
					print fCMD ";\n";
					}
			print fCMD "\t\tsavetrees file=finalRatchet.tre format=nexus replace;\n";
			print fCMD "\n\t[!][!*** Determining consensus trees ***]\n\n";
			print fCMD "\t\tcontree / strict=yes majrule=yes indices=yes grpfreq=yes treefile=conRatchet.tre replace;\n";
			}
		else
			{
			if ($NumBremer > 1)
				{
				print fCMD "\t\tlog file=BremerRatchet.log ";
				if ($TreeRepl == 1)
					{
					print fCMD "replace;\n";
					}
					else
					{
					print fCMD "append;\n";
					}
				print fCMD "\t\t\tgettrees file=ratchet.tre allblocks=yes warntree=no;\n";
				print fCMD "\t\t\tfilter best=yes;\n";
				print fCMD "\t\t\tdescribetrees / plot=none;\n";
				print fCMD "\t\tlog stop;\n\n";
				}
			else
				{
				print fCMD "\t[!][!*** Determining consensus trees ***]\n\n";
				print fCMD "\tgettrees file=ratchet.tre allblocks=yes warntree=no;\n";
				print fCMD "\t\tfilter best=yes;\n";
				print fCMD "\t\tdescribetrees / plot=none;\n";
				print fCMD "\t\tcontree / strict=yes majrule=yes indices=yes grpfreq=yes treefile=conRatchet.tre replace;\n";
				}
			}
		}

	print fCMD "\n\tlog stop;\n" unless ($NumBremer > 1);
	print fCMD "end;\n\n";
	print fCMD "quit warntsave=no;\n";
close fCMD;

# Subroutines used in the script

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}
